/*
    comp4
    
    This program is a syntactic analyzer and evaluator for the following grammar:
    
       Expr::=  '(' oper Expr  Expr ')' | number
       oper ::= '+' | '-' 
       number ::= '0'| '1' | ... | '9'
       
    The evaluation of the expression is printed in the standard output     
*/

public class Compiler {

    public int compile( char []p_input ) {
        input = p_input;
        tokenPos = 0;
        nextToken();
        int result = expr();
        if ( token != '\0' )
          error();
        return result;
    }
    
    private int expr() {
        if ( token == '(' ) {
            nextToken();
            char op = oper();
            int left  = expr();
            int right = expr();
            int result = 0;
            switch (op) {
                case '+' :
                  result = left + right;
                  break;
                case '-' :
                  result = left - right;
                  break;
            }
            if ( token == ')' ) 
              nextToken();
            else
              error();
            return result;
        }
        else
          return number();
    }
    
    
    private int number() {
        int result = 0;
        
        if ( token >= '0'  && token <= '9' ) {
            result = token - '0';
            nextToken();
        }
        else
          error();
        return result;
    }
    
    private char oper() {
        char op = token;
        if ( token == '+' || token == '-' ) {
          nextToken();
          return op;
        }
        else {
          error();
            // should never execute since error terminates the program
          return '\0';
        }
    }
    
    private void nextToken() {
        while (  tokenPos < input.length && input[tokenPos] == ' ' ) 
          tokenPos++;
        if ( tokenPos < input.length ) {
          token = input[tokenPos];        
          tokenPos++;
        }
        else
          token = '\0';
    }
    
    private void error() {
        if ( tokenPos == 0 ) 
          tokenPos = 1; 
        else 
          if ( tokenPos >= input.length )
            tokenPos = input.length;
        
        String strInput = new String( input, tokenPos - 1, input.length - tokenPos + 1 );
        String strError = "Error at \"" + strInput + "\"";
        System.out.print( strError );
        throw new RuntimeException(strError);
    }
    
    private char token;
    private int  tokenPos;
    private char []input;
    
}
    
    
    