/*
    comp5
    
    This program is a syntactic analyzer that builds the abstract syntax tree (AST). The 
    classes of the AST were put in the package AST. The classes of the AST represent  
    expressions and all have a "genC" method. If exp is a variable of class Expr, then
          exp.genC()
    generates code in C for the expression.
    
    Grammar:
    
       Expr::=  '(' oper Expr  Expr ')' | number
       oper ::= '+' | '-' 
       number ::= '0'| '1' | ... | '9'
       
     
*/

import AST.*;

public class Compiler {


    public Expr compile( char []p_input ) {
        input = p_input;
        tokenPos = 0;
        nextToken();
        Expr e = expr();
        if ( token != '\0' )
          error();
        return e;
    }
    
    private Expr expr() {
        if ( token == '(' ) {
            nextToken();
            char op = oper();
            Expr e1 = expr();
            Expr e2 = expr();
            CompositeExpr ce = new CompositeExpr(e1, op, e2);
            if ( token == ')' ) 
              nextToken();
            else
              error();
            return ce;
        }
        else
          return number();
    }
    
    private NumberExpr number() {
        NumberExpr e = null;
        
        if ( token >= '0'  && token <= '9' ) {
            e = new NumberExpr(token);    
            nextToken();
        }
        else
          error();
        return e;
    }
    
    private char oper() {
        char op = token;
        if ( token == '+' || token == '-' ) {
          nextToken();
          return op;
        }
        else {
          error();
            // should never execute since error terminates the program
          return '\0';
        }
    }
    
    private void nextToken() {
        while (  tokenPos < input.length && input[tokenPos] == ' ' ) 
          tokenPos++;
        if ( tokenPos < input.length ) {
          token = input[tokenPos];        
          tokenPos++;
        }
        else
          token = '\0';
    }
    
    private void error() {
        if ( tokenPos == 0 ) 
          tokenPos = 1; 
        else 
          if ( tokenPos >= input.length )
            tokenPos = input.length;
        
        String strInput = new String( input, tokenPos - 1, input.length - tokenPos + 1 );
        String strError = "Error at \"" + strInput + "\"";
        System.out.print( strError );
        throw new RuntimeException(strError);
    }
    
    private char token;
    private int  tokenPos;
    private char []input;

}
    
    
    