/*
    comp6
    
    
    We added to the grammar more operators and a declaration of variables so that 
    the grammar now accepts a program like 
          a = 1  b = 3  : (- (+ a 2) 3)
    The result of the evaluation would be 0. Another program would be
          g = 3  t = 9  : (* (- t 7) g)

    Of course, the AST was extended to cope with the new rules. New classes
    were created: 
         Program - represents the program
         Variable - a variable in the declaration
         VariableExpr - a variable inside an expression
         
    Grammar:
       Program ::= VarDecList ':' Expr   
       VarDecList ::=  | VarDec VarDecList 
       VarDec ::= Letter '=' Number
       Expr::=  '(' oper Expr  Expr ')' | Number | Letter
       Oper ::= '+' | '-' | '*' | '/'
       Number ::= '0'| '1' | ... | '9'
       Letter ::= 'A' | 'B'| ... | 'Z'| 'a'| 'b' | ... | 'z'
       
     
*/

import AST.*;
import java.util.Vector;
import java.lang.Character;

public class Compiler {

    public Program compile( char []p_input ) {
        input = p_input;
        tokenPos = 0;
        nextToken();
        Program result = program();
        if ( token != '\0' )
          error();
        return result;
        }
    
    private Program program() {
        Vector arrayVariable = varDecList();
        if ( token != ':' ) {
            error();
            return null;
        }
        else {
            nextToken();
            Expr e = expr();
            return new Program( arrayVariable, e );
        }
    }

    private Vector varDecList() {
          /* See how the repetition in the grammar reflects in the code. Since VarDec
             always begin with a letter, if token is NOT a letter, then VarDecList is
             empty and null is returned */
        if ( ! Character.isLetter(token) )
          return null;
        else {
            Vector arrayVariable = new Vector();
            while ( Character.isLetter(token) ) 
              arrayVariable.addElement( varDec() );
            return arrayVariable;
            }
            
    }
    
    
    private Variable varDec() {
        char name = letter();
        if ( token != '=' ) {
          error();
          return null;
        }
        else {
          nextToken();
          NumberExpr n = number();
          return new Variable( name, n.getValue() );
        }
    }

    private char letter() {
       if ( ! Character.isLetter(token) ) {
          error();
          return '\0';
       }
       else {
         char ch = token;
         nextToken();
         return ch;
       }
    }
    
    private Expr expr() {
        if ( token == '(' ) {
            nextToken();
            char op = oper();
            Expr e1 = expr();
            Expr e2 = expr();
            CompositeExpr ce = new CompositeExpr(e1, op, e2);
            if ( token == ')' ) 
              nextToken();
            else
              error();
            return ce;
        }
        else
            // note we test the token to decide which production to use
          if ( Character.isDigit(token) ) 
            return number();
          else
            return new VariableExpr(letter());
    }
    
    private NumberExpr number() {
        NumberExpr e = null;
        
        if ( token >= '0'  && token <= '9' ) {
            e = new NumberExpr(token);    
            nextToken();
        }
        else
          error();
        return e;
    }
    
    private char oper() {
        char op = token;
        if ( token == '+' || token == '-' || token == '*'  || token == '/' ) {
          nextToken();
          return op;
        }
        else {
          error();
            // should never execute since error terminates the program
          return '\0';
        }
    }
    
    private void nextToken() {
        while (  tokenPos < input.length && input[tokenPos] == ' ' ) 
          tokenPos++;
        if ( tokenPos < input.length ) {
          token = input[tokenPos];        
          tokenPos++;
        }
        else
          token = '\0';
    }
    
    private void error() {
        if ( tokenPos == 0 ) 
          tokenPos = 1; 
        else 
          if ( tokenPos >= input.length )
            tokenPos = input.length;
        
        String strInput = new String( input, tokenPos - 1, input.length - tokenPos + 1 );
        String strError = "Error at \"" + strInput + "\"";
        System.out.print( strError );
        throw new RuntimeException(strError);
    }
    
    private char token;
    private int  tokenPos;
    private char []input;
    
}
