/*
    comp11
    
    This compiler prints the names of the methods called during 
    syntax analysis.
    
*/

import AST.*;
import java.util.*;
import java.lang.Character;
import Lexer.*;
import java.io.*;


public class Compiler {
    
      // compile must receive an input with an character less than 
      // p_input.lenght
    public Program compile( char []input, PrintWriter outError, 
                            PW pw ) {
        
        symbolTable = new Hashtable();
        error = new CompilerError( outError );
        lexer = new Lexer(input, error);
        error.setLexer(lexer);
        
        lexer.nextToken();
        return program(pw);
        }
    
    private Program program(PW pw) {
        // Program ::= [ "var" VarDecList ] CompositeStatement

        pw.println("/*");
        pw.add();
        pw.println("program");
        
        Vector arrayVariable = null;
        
        if ( lexer.token == Symbol.VAR )  {
          lexer.nextToken();
          arrayVariable = varDecList(pw);
        }
        
        StatementList sl = compositeStatement(pw);
        
        pw.println("new Program(?,?) in method program");
        
        Program program = new Program( arrayVariable, sl );
        if ( lexer.token != Symbol.EOF ) 
          error.signal("EOF expected");
          
        pw.sub();
        pw.println("*/");
        pw.println("");
        pw.println("");
        return program;
    }
        

    private StatementList compositeStatement(PW pw) {
          // CompositeStatement ::= "begin" StatementList "end"
          //  StatementList ::= | Statement ";" StatementList
        
        pw.add();
        pw.println("compositeStatement");
        
        if ( lexer.token != Symbol.BEGIN ) 
            error.signal("BEGIN expected");
        lexer.nextToken();
        StatementList sl = statementList(pw);
        if ( lexer.token != Symbol.END )
          error.signal("\"end\" expected");
        lexer.nextToken();
        
        pw.sub();
        
        return sl;
    }

    private StatementList statementList(PW pw) {
        
        pw.add();
        pw.println("statementList");
        
        Vector v = new Vector();
          // statements always begin with an identifier, if, read or write
        while ( lexer.token == Symbol.IDENT ||
                lexer.token == Symbol.IF ||
                lexer.token == Symbol.READ ||
                lexer.token == Symbol.WRITE ) {
            v.addElement( statement(pw) );
            if ( lexer.token != Symbol.SEMICOLON )
              error.signal("; expected");
            lexer.nextToken();
            }
            
        pw.println("new StatementList(?) in method statementList");
        pw.sub();
        
        return new StatementList(v);
    }
            
            
    private Statement statement(PW pw) {
        /*  Statement ::= AssignmentStatement | IfStatement | ReadStatement |
                          WriteStatement
        */
        
        pw.add();
        pw.println("statement");
        
        Statement ret = null;
        switch (lexer.token) {
            case Symbol.IDENT :
              ret = assignmentStatement(pw);
              break;
            case Symbol.IF :
              ret = ifStatement(pw);
              break;
            case Symbol.READ :
              ret = readStatement(pw);
              break;
            case Symbol.WRITE :
              ret = writeStatement(pw);
              break;
            default :
                // will never be executed
              error.signal("Statement expected");
        }
        
        pw.sub();
        
        return ret;
    }
    
    
    
    private AssignmentStatement assignmentStatement(PW pw) {
        
        
        pw.add();
        pw.println("assignmentStatement");
        
        String name = (String ) lexer.getStringValue();
        
          // is the variable in the symbol table ? Variables are inserted in the
          // symbol table when they are declared. It the variable is not there, it has
          // not been declared.
          
        pw.println("Search for \"" + name + "\" in the symbol table");  
          
        Variable v = (Variable ) symbolTable.get(name);
          // was it in the symbol table ?
        if ( v == null ) 
            error.signal("Variable " + name + " was not declared");
        lexer.nextToken();
        if ( lexer.token != Symbol.ASSIGN ) 
          error.signal("= expected");
        lexer.nextToken();
        Expr right = orExpr(pw);
          // semantic analysis
          // check if expression has the same type as variable
        if ( v.getType() != right.getType() )
          error.signal("Type error in assignment");
        
        pw.println("new AssignmentStatement(" + name + ", ?) in method assignmentStatement");
        pw.sub();
        
        return new AssignmentStatement( v, right );
    }
    
    
    private IfStatement ifStatement(PW pw) {
        
        pw.add();
        pw.println("ifStatement");
        
        lexer.nextToken();
        Expr e = orExpr(pw);
          // semantic analysis
          // check if expression has type boolean
        if ( e.getType() != Type.booleanType ) 
          error.signal("Boolean type expected in if expression");
        
        if ( lexer.token != Symbol.THEN ) 
          error.signal("then expected");
        lexer.nextToken();
        StatementList thenPart = statementList(pw);
        StatementList elsePart = null;
        if ( lexer.token == Symbol.ELSE ) {
            lexer.nextToken();
            elsePart = statementList(pw);
        }
        if ( lexer.token != Symbol.ENDIF )
          error.signal("\"endif\" expected");
        lexer.nextToken();
        
        pw.println("new IfStatement(?, ?, ?) in method ifStatement");
        pw.sub();
        
        return new IfStatement( e, thenPart, elsePart );
    }
            
          
            
    private ReadStatement readStatement(PW pw) {
        
        pw.add();
        pw.println("readStatement");
        
        lexer.nextToken();
        if ( lexer.token != Symbol.LEFTPAR ) 
          error.signal("( expected");
        lexer.nextToken();
        if ( lexer.token != Symbol.IDENT )
          error.signal("Identifier expected");
          // semantic analysis
          // check if the variable was declared
        String name = (String ) lexer.getStringValue();
        
        pw.println("Search for \"" + name + "\" in the symbol table");  
        
        Variable v = (Variable ) symbolTable.get(name);
        if ( v == null ) 
          error.signal("Variable " + name + " was not declared");
          // semantic analysis
          // check if variable has type char or integer
        if ( v.getType() != Type.charType && v.getType() != Type.integerType ) 
          error.signal("Variable should have type char or integer");
          
        lexer.nextToken();
        if ( lexer.token != Symbol.RIGHTPAR ) 
          error.signal(") expected");
        lexer.nextToken();
        
        pw.println("new ReadStatement(" + name + ") in method readStatement");
        pw.sub();
        
        return new ReadStatement( v );
    }
    
    
    private WriteStatement writeStatement(PW pw) {
        
        pw.add();
        pw.println("writeStatement");
        
        lexer.nextToken();
        if ( lexer.token != Symbol.LEFTPAR ) 
          error.signal("( expected");
        lexer.nextToken();
          // expression may be of any type
        Expr e = orExpr(pw);
        if ( lexer.token != Symbol.RIGHTPAR ) 
          error.signal(") expected");
        lexer.nextToken();
        
        pw.println("new WriteStatement(?) in method writeStatement");
        pw.sub();
        
        return new WriteStatement( e );
    }
        
        
            
            
    private Vector varDecList(PW pw) {
        //  VarDecList ::= VarDecList2  { VarDecList2  }

        pw.add();
        pw.println("varDecList");
        
        Vector varList = new Vector();
        
        varDecList2(varList, pw);
        while ( lexer.token == Symbol.IDENT )
          varDecList2(varList, pw);
          
        pw.sub();
        return varList;
    }

    private void varDecList2( Vector varList, PW pw ) {
        //  VarDecList2 ::= Ident { ',' Ident } ':' Type ';'

        pw.add();
        pw.println("varDecList2");

        Vector lastVarList = new Vector();
        
        while ( true ) {
          if ( lexer.token != Symbol.IDENT )
            error.signal("Identifier expected");
            // name of the identifier
          String name = lexer.getStringValue();
          lexer.nextToken();
            // semantic analysis
            // if the name is in the symbol table, the variable is been declared twice.
            
          pw.println("Search for \"" + name + "\" in the symbol table");  
            
          if ( symbolTable.get(name) != null )
            error.signal("Variable " + name + " has already been declared");

          pw.println("new Variable(" + name + ") in method varDecList2");

            // variable does not have a type yet
          Variable v = new Variable(name);
            // inserts the variable in the symbol table. The name is the key and an
            // object of class Variable is the value. Hash tables store a pair (key, value)
            // retrieved by the key.
            
          pw.println("Variable " + name + " inserted in the symbol table");  
            
          symbolTable.put( name, v );
            // list of the last variables declared. They don't have types yet
          lastVarList.addElement(v);
          
          if ( lexer.token == Symbol.COMMA ) 
            lexer.nextToken();
          else 
            break;
        }

        if ( lexer.token != Symbol.COLON ) //  :
          error.signal(": expected");
        lexer.nextToken();
          // get the type
        Type typeVar = type(pw);
        
        Variable v;
        Enumeration e = lastVarList.elements();
        while ( e.hasMoreElements() ) {
            // add type to the variable
          v = (Variable ) e.nextElement();
          v.setType(typeVar);
            // add variable to the list of variable 
          varList.addElement(v);
        }
        
        if ( lexer.token != Symbol.SEMICOLON ) 
          error.signal("; expected");
        lexer.nextToken();  
        
        pw.sub();
    }
          
        

    private Type type(PW pw) {
        
        pw.add();
        pw.println("type");
        
        Type result;
        
        switch ( lexer.token ) {
            case Symbol.INTEGER :
              result = Type.integerType;
              break;
            case Symbol.BOOLEAN :
              result = Type.booleanType;
              break;
            case Symbol.CHAR :
              result = Type.charType;
              break;
            default :
              error.signal("Type expected");
              result = null;
        }
        lexer.nextToken();
        
        pw.sub();
        
        return result;
        
    }
    

    private Expr orExpr(PW pw) {
      /*  
        OrExpr ::= AndExpr [ "or" AndExpr ]
      */
      
      pw.add();
      pw.println("orExpr");
      
      Expr left, right;
      left = andExpr(pw);
      if ( lexer.token == Symbol.OR ) {
        lexer.nextToken();
        right = andExpr(pw);
          // semantic analysis
        if ( left.getType() != Type.booleanType ||
             right.getType() != Type.booleanType )
           error.signal("Expression of boolean type expected");
           
        pw.println("new CompositeExpr(?, ?, ?) in method orExpr");
        
        left = new CompositeExpr(left, Symbol.OR, right);
      }
      
      pw.sub();
      
      return left;
    }

    private Expr andExpr(PW pw) {
      /*
        AndExpr ::= RelExpr [ "and" RelExpr ]
      */
      
      pw.add();
      pw.println("andExpr");
      
      Expr left, right;
      left = relExpr(pw);
      if ( lexer.token == Symbol.AND ) {
        lexer.nextToken();
        right = relExpr(pw);
          // semantic analysis
        if ( left.getType() != Type.booleanType ||
             right.getType() != Type.booleanType )
           error.signal("Expression of boolean type expected");
           
        pw.println("new CompositeExpr(?, ?, ?) in method andExpr");
           
        left = new CompositeExpr( left, Symbol.AND, right );
      }
      
      pw.sub();
      
      return left;
    }

    private Expr relExpr(PW pw) {
      /*
        RelExpr ::= AddExpr [ RelOp AddExpr ]
      */
      
      pw.add();
      pw.println("relExpr");
      
      Expr left, right;
      left = addExpr(pw);
      int op = lexer.token;
      if ( op == Symbol.EQ || op == Symbol.NEQ || op == Symbol.LE || op == Symbol.LT ||
           op == Symbol.GE || op == Symbol.GT  ) {
         lexer.nextToken();
         right = addExpr(pw);
           // semantic analysis
         if ( left.getType() != right.getType() ) 
           error.signal("Type error in expression");
           
         pw.println("new CompositeExpr(?, ?, ?) in method relExpr");
           
         left = new CompositeExpr( left, op, right );
       }
       
       pw.sub();
       
       return left;
    }
    
    private Expr addExpr(PW pw) {
      /*
        AddExpr ::= MultExpr { AddOp MultExpr }
        
      */
      
      pw.add();
      pw.println("addExpr");
      
      int op;
      Expr left, right;
      left = multExpr(pw);
      while ( (op = lexer.token) == Symbol.PLUS ||
              op == Symbol.MINUS ) {
        lexer.nextToken();
        right = multExpr(pw);
          // semantic analysis
        if ( left.getType() != Type.integerType ||
             right.getType() != Type.integerType )
           error.signal("Expression of type integer expected");
           
        pw.println("new CompositeExpr(?, ?, ?) in method addExpr");
           
        left = new CompositeExpr( left, op, right );
      }
      
      pw.sub();
      
      return left;
    }
       
    private Expr multExpr(PW pw) {
     /*
        MultExpr ::= SimpleExpr { MultOp SimpleExpr }
     */    
     
       pw.add();
       pw.println("multExpr");
     
       Expr left, right;
       left = simpleExpr(pw);
       int op;
       while ( (op = lexer.token) == Symbol.MULT ||
                op == Symbol.DIV || op == Symbol.REMAINDER ) {
          lexer.nextToken();
          right = simpleExpr(pw);
            // semantic analysis
          if ( left.getType() != Type.integerType ||
             right.getType() != Type.integerType )
            error.signal("Expression of type integer expected");

          pw.println("new CompositeExpr(?, ?, ?) in method multExpr");
          
          left = new CompositeExpr( left, op, right );
       }
       
       pw.sub();
       
       return left;
    }
           
    private Expr simpleExpr(PW pw) {
      /*
        SimpleExpr ::= Number |  "true" | "false" | Character 
           | '(' Expr ')' | "not" SimpleExpr | Variable 
      */
      
      pw.add();
      pw.println("simpleExpr");
      
      Expr e, ret;
      
        // note we test the lexer.getToken() to decide which production to use
      switch ( lexer.token ) {
         case Symbol.NUMBER :
            ret = number(pw);
            break;
         case Symbol.TRUE :
           lexer.nextToken();
           ret = BooleanExpr.True;
           break;
         case Symbol.FALSE :
           lexer.nextToken();
           ret = BooleanExpr.False;
           break;
         case Symbol.CHARACTER :
             // get the token with getToken.
             // then get the value of it, with has the type Object
             // convert the object to type Character using a cast
             // call method charValue to get the character inside the object
           char ch = lexer.getCharValue();
           lexer.nextToken();
           
           pw.println("new CharExpr(" + ch + ") in method simpleExpr");
           
           ret = new CharExpr(ch);
           break;
         case Symbol.LEFTPAR :
           lexer.nextToken();
           ret = orExpr(pw);
           if ( lexer.token != Symbol.RIGHTPAR )
             error.signal(") expected");
           lexer.nextToken();
           break;
         case Symbol.NOT :
           lexer.nextToken();
           e = orExpr(pw);
             // semantic analysis
           if ( e.getType() != Type.booleanType )
             error.signal("Expression of type boolean expected");
             
           pw.println("new UnaryExpr(?, Symbol.NOT) in method simpleExpr");
             
           ret = new UnaryExpr( e, Symbol.NOT );
           break;
         case Symbol.PLUS :
           lexer.nextToken();
           e = orExpr(pw);
             // semantic analysis
           if ( e.getType() != Type.integerType )
             error.signal("Expression of type integer expected");
             
           pw.println("new UnaryExpr(?, Symbol,PLUS) in method simpleExpr");

           ret = new UnaryExpr( e, Symbol.PLUS );
           break;
         case Symbol.MINUS :
           lexer.nextToken();
           e = orExpr(pw);
             // semantic analysis
           if ( e.getType() != Type.integerType )
             error.signal("Expression of type integer expected");
             
           pw.println("new UnaryExpr(?, Symbol.MINUS) in method simpleExpr");
             
           ret = new UnaryExpr( e, Symbol.MINUS );
           break;
         default :
             // an identifier
           if ( lexer.token != Symbol.IDENT ) 
             error.signal("Identifier expected");
           String name = (String ) lexer.getStringValue();
           lexer.nextToken();
           
           pw.println("Search for \"" + name + "\" in the symbol table");  
           
           Variable v = (Variable ) symbolTable.get( name );
             // semantic analysis
             // was the variable declared ? 
           if ( v == null ) 
             error.signal("Variable " + name + " was not declared");
             
           pw.println("new VariableExpr(?) in method simpleExpr");
           
           ret = new VariableExpr(v);
        }
        
      pw.sub();
      
      return ret;
            
    }
    
    
    
    private NumberExpr number(PW pw) {
        
        pw.add();
        pw.println("number");
        
        NumberExpr e = null;
        
          // the number value is stored in lexer.getToken().value as an object of Integer.
          // Method intValue returns that value as an value of type int.
        int value = lexer.getNumberValue();
        lexer.nextToken();
        
        pw.println("new NumberExpr(" + value + ") in method number");
        pw.sub();
        
        return new NumberExpr( value );
    }
    
    
    private boolean checkTypes( Type left, int op, Type right ) {
        // return true if  left and right can be the types of a composite
        // expression with operator op 

       if ( op == Symbol.EQ || op == Symbol.NEQ || op == Symbol.LE || op == Symbol.LT ||
                 op == Symbol.GE || op == Symbol.GT )
          return left == right;
       else if ( op == Symbol.PLUS || op == Symbol.MINUS || op == Symbol.DIV ||
                 op == Symbol.MULT || op == Symbol.REMAINDER ) {
                
          if ( left != right || left != Type.integerType )
             return false;
          else 
             return true;
       }
       else if ( op == Symbol.AND || op == Symbol.OR ) {
          if ( left != Type.booleanType || 
               right != Type.booleanType )
             return false;
          else 
             return true;
       }
       else {
          error.signal("Compiler internal error: unknown operator");
          return true;
       }
    }
    
    

    
    
    private Hashtable symbolTable;
    private Lexer lexer;
    private CompilerError error;

        
}
