import java_cup.runtime.*;
import java.lang.*;


%%

  // what to return at EOF ? Return symbol EOF
%eofval{
{
  return new Symbol(Sym.EOF, yychar, yychar + yylength(), null);
}
%eofval}       

  // the code below will be placed inside the class Lexer
%{
    
   public int getLineNumber() {
      return yyline;
   }
%}
   

%char
%line

  // the following three lines give the name Interface the generated class will implement,
  // the name of the method for getting the next token and the type of the tokens.
  // they can be replaced by "%cup"
%implements java_cup.runtime.Scanner
%function next_token
%type java_cup.runtime.Symbol
  // make class Lexer a public class
%public
  // the name of the generated class should be "Lexer"
%class Lexer

  // The symbols MINUS, PLUS, ... are declared in class Sym generated by CUP. See
  // option -symbols of CUP.
%%

"-"             { return new Symbol( Sym.MINUS ); }
"+"             { return new Symbol( Sym.PLUS ); }
"*"             { return new Symbol( Sym.MULT ); }
"/"             { return new Symbol( Sym.DIV ); }
"<="            { return new Symbol( Sym.LE ); }
"<"             { return new Symbol( Sym.LT ); }
">="            { return new Symbol( Sym.GE ); }
">"             { return new Symbol( Sym.GT ); }
"=="            { return new Symbol( Sym.EQ ); }
"<>"            { return new Symbol( Sym.NEQ ); }
"("             { return new Symbol( Sym.LEFTPAR, null ); }
")"             { return new Symbol( Sym.RIGHTPAR, null ); }
[0-9][0-9]*     { return new Symbol( Sym.NUMBER, new Integer(yytext()) ); }


"="             { return new Symbol( Sym.ASSIGN ); }
","             { return new Symbol( Sym.COMMA ); }
";"             { return new Symbol( Sym.SEMICOLON); }
"var"           { return new Symbol( Sym.VAR ); }
"begin"         { return new Symbol( Sym.BEGIN ); }
"end"           { return new Symbol( Sym.END ); }
"if"            { return new Symbol( Sym.IF); }
"then"          { return new Symbol( Sym.THEN); }
"else"          { return new Symbol( Sym.ELSE); }
"endif"         { return new Symbol( Sym.ENDIF); }
"read"          { return new Symbol( Sym.READ); }
"write"         { return new Symbol( Sym.WRITE); }

[A-Za-z][A-Za-z]*   { return new Symbol( Sym.IDENT, yytext() ); }

[ \t\r\n]       {}
. { System.out.println("unmatched: \"" + yytext() + "\"" + yyline); }

