import java_cup.runtime.*;
import AST.*;
import java.io.*;

public class Main {




  public static void main(String args[]) throws IOException {
    
      int i = 0;
      if ( args.length < 1 ) {
        System.out.println("Program needs a parameter");
        return;
      }
        // input is from a file
      String filename = args[0];
      

        // never do what is done below. First we read the file to variable input and
        // then we create a "input stream" with the file and pass it to the lexer.
        // the correct would be to create a class CharInputStream whose constructor
        // would take variable input as parameter. Then we could pass a variable of
        // this class to the lexer.
      File file = new File(args[0]);
      FileReader stream;
      if ( ! file.exists() || ! file.canRead() ) {
         System.out.println("Either the file " + args[0] + " does not exist or it cannot be read");
         throw new RuntimeException();
      }
      try { 
         stream = new FileReader(file);  
      } catch ( FileNotFoundException e ) {
         System.out.println("Something wrong: file does not exist anymore");
         throw new RuntimeException();
      }
         // one more character for '\0' at the end that will be added by the
         // compiler
      char []input = new char[ (int ) file.length() + 1 ];
      int numChRead;
            
      try {
         numChRead = stream.read( input, 0, (int ) file.length() );
      } catch ( IOException e ) {
         System.out.println("Error reading file " + args[0]);
         throw new RuntimeException();
      }
                
      if ( numChRead != file.length() ) {
         System.out.println("Read error");
         throw new RuntimeException();
      }
      try {
         stream.close();
      } catch ( IOException e ) {
         System.out.println("Error in handling the file " + args[0]);
         throw new RuntimeException();
      }
                

      
      
      
      InputStream inp = new FileInputStream(filename);
      Lexer lexer = new Lexer(inp);
      
      
        /* create a parsing object */
      Parser parser_obj = new Parser(lexer, input);

      try {
            // call the parser. The result will be an object Symbol whose instance
            // variable value points to an object with the AST
          Program p = (Program ) parser_obj.parse().value;
          p.genC();
      } catch (Exception e) {
         System.out.println(e);
         e.printStackTrace();
      } 
  }
}



        